/********************************************************************************/
/*                    VMware BIOS Injection Code Example                        */
/*          Copyright (C) 2023 Daniel B. Sedory - Public Domain                 */
/********************************************************************************/
/*  Serial output access data and code was provided by Pete Batard from here:   */
/*          https://pete.akeo.ie/2011/06/crafting-bios-from-scratch.html        */
/********************************************************************************/

/********************************************************************************/
/* GNU Assembler Settings:                                                      */
/********************************************************************************/
.intel_syntax noprefix	/*     Use Intel assembler syntax (same as MASM)        */
.code16			/* The VMWare (Phoenix) BIOS BB.ROM code; at least      */
                        /* most of it, begins executing in 16-bit Real Mode     */
                        /* as does all x86 CPU BIOS code!                       */
/********************************************************************************/
/* Macro:                                                                       */
/********************************************************************************/
/* This GNU AS Macro makes coding stackless subroutine calls much easier:       */

.macro  ROM_CALL addr
	mov  sp, offset 1f	/* Use a local label as we don't know the size  */
	jmp  \addr		/* of the jmp instruction (can be 2 or 3 bytes) */
1:
.endm
        /* See: https://sourceware.org/binutils/docs-2.21/as/Symbol-Names.html  */

/********************************************************************************/
/* Constants:                                                                   */
/********************************************************************************/
/*   The VMware platform uses an emulated NS 338 Super I/O chipset; which       */
/*    is compliant with the PC97338 Super I/O data found on the Internet.       */

SUPERIO_BASE  = 0x2e	/* The VMware PC97338 Super I/O Base is: 0x2E           */
			/* (Some code in the BIOS does attempt to communicate   */
			/*  via Port 0x398, but doesn't appear to be used for   */
			/*  Serial Port communications.)                        */

PC97338_FER   = 0x00	/* PC97338 Function Enable Register                     */
PC97338_FAR   = 0x01	/* PC97338 Function Address Register                    */
PC97338_PTR   = 0x02	/* PC97338 Power and Test Register                      */

/* 16650 UART Setup                                                             */
/* ----------------                                                             */
COM_BASE      = 0x3f8	/* Our default COM1 base, after SuperIO init            */
COM_RB        = 0x00	/* Receive Buffer (R)                                   */
COM_TB        = 0x00    /* Transmit Buffer (W)                                  */
COM_BRD_LO    = 0x00	/* Baud Rate Divisor LSB (when bit 7 of LCR is set)     */
COM_BRD_HI    = 0x01	/* Daud Rate Divisor MSB (when bit 7 of LCR is set)     */
COM_IER       = 0x01	/* Interrupt Enable Register                            */
COM_FCR       = 0x02	/* 16650 FIFO Control Register (W)                      */
COM_LCR       = 0x03	/* Line Control Register                                */
COM_MCR       = 0x04	/* Modem Control Registrer                              */
COM_LSR       = 0x05	/* Line Status Register                                 */
/********************************************************************************/


/********************************************************************************/
/* begin : Dummy section marking the very start of the BIOS.                    */
/* This allows the .rom binary to be filled to the right size with objcopy.     */
/********************************************************************************/
.section begin, "a"		/* The 'ALLOC' flag is needed for objcopy       */
	.ascii "VMBIOS v1.30"	/* Dummy ID string                              */
	.align 16
/********************************************************************************/


/********************************************************************************/
/* Section marking the start of the Serial Output code to inject into BIOS.     */
/*                                                                              */
/* main:                                                                        */
/*         This section will be relocated according to our serial.ld script.    */
/********************************************************************************/

.section main, "ax"
.globl init		/* init must be declared global for the linker and must */
			/* point to the first instruction of our CODE section.  */

init:
	xor	al, 0x80	/* Sets only high 'nibble' (high 4 bits of AL)  */
				/*   to zero (only if, AL is 80h to 8Fh; which  */
				/*   I already discovered it would be!).        */
	or	al, 0x30	/* Turns AL into ASCII equivalent for 0 - 9,    */
				/*   but would output the characters:           */
				/*   :   ;   <   =   >  ?   for 0A - 0F.        */

	mov	di, ax		/* Store converted AL byte in lower half of     */
				/* DI Register; ignoring whatever is in AH.     */
	mov	ax, cs
	mov	ds, ax	      /* Data Segment must be the same as Code Segment. */


/*	cli		/* Shouldn't be necessary; we're already in BIOS code.  */
	cld		/*       Set String manipulation direction to Forward.  */

init_superio:
	mov  dx, SUPERIO_BASE	/*  The PC97338 datasheet says we are supposed  */
	in   al, dx		/* to read this port twice on startup, but the  */
	in   al, dx		/* VMware virtual chip doesn't seem to care...  */

	/* Send Super I/O chipset Configuration bytes from our Data Section:    */

	mov  si, offset superio_conf
	mov  cx, (serial_conf - superio_conf)/2

write_superio_conf:
	mov  ax, [si]
	ROM_CALL superio_out

	add  si, 0x02
	loop write_superio_conf

init_serial:	/* Initialize Serial Port using Serial Configuration bytes:     */

	mov  si, offset serial_conf
	mov  cx, (msg1 - serial_conf)/2

write_serial_conf:
	mov  ax, [si]
	ROM_CALL serial_out

	add  si, 0x02
	loop write_serial_conf

	/* Send our first ASCII string out the Serial Port. */

	mov  si, offset msg1
	ROM_CALL print_string		/* Print msg1 string.                   */

	mov	ax, di		/* Copy converted Diagnostic Code back into AL. */

	mov	dx, COM_BASE + COM_LSR
	mov	ah, al	    /* Store [AL] in AH, because AL must be used below! */

tx_stall:
	in	al, dx
	and	al, 0x20	/*  Check that transmit register is empty.      */
	jz	tx_stall

	mov	dx, COM_BASE + COM_TB
	mov	al, ah		/* Copy [AH] back into AL, and ...              */
	out	dx, al		/*   send the "Code byte" out Serial Port.      */

	mov	si, offset msg2		/* Finish print line with msg2 string.  */
	ROM_CALL print_string		/* Print msg2 string.                   */

	jmp	0xE756		/* Return to 'hardcoded' address in the...      */
				/*    official VMware BIOS code!                */


/********************************************************************************/
/* Subroutines:                                                                 */
/********************************************************************************/

superio_out:			/*    AL: Register Index,  AH: Data to Write.   */
	mov  dx, SUPERIO_BASE
	out  dx, al
	inc  dx
	xchg al, ah
	out  dx, al
	jmp  sp

serial_out:			/* AL: COM Register Index, AH: Data to Write.   */
	mov  dx, COM_BASE
	add  dl, al	/* Unless something is wrong, we won't overflow to DH.  */
	mov  al, ah
	out  dx, al
	jmp  sp

putchar:				/*      AL: Character to Print.         */
	mov  dx, COM_BASE + COM_LSR
	mov  ah, al

tx_wait:
	in   al, dx
	and  al, 0x20		/*   Check that transmit register is empty.     */
	jz   tx_wait
	mov  dx, COM_BASE + COM_TB
	mov  al, ah
	out  dx, al
	jmp  sp

print_string:		/* SI (for LODSB): Offset to NUL terminated string.     */

	lodsb			/* Load String Bytes from SI pointer.           */
	or   al, al		/* Will keep sending characters until ...       */
	jnz  write_char		/* NUL (zero byte) is encountered.              */
	jmp  sp

write_char:
	shl  esp, 0x10	/* Possibly the only way left to preserve the value     */
			/* currently in the SP register, since all Registers    */
  /* are in use, _but_ we must call another subroutine! So we'll store this     */
  /* SP value in the high 16 bits of the 32-bit ESP Register.                   */

	ROM_CALL putchar

	shr  esp, 0x10		/* Restore previously saved SP value            */
				/*   from high 16 bits of ESP.                  */
	jmp  print_string


/********************************************************************************/
/* Data:                                                                        */
/********************************************************************************/

superio_conf:
/* http://www.datasheetcatalog.org/datasheet/nationalsemiconductor/PC97338.pdf  */

	.byte PC97338_FER, 0x0f		/* Enable COM, PAR and FDC              */
	.byte PC97338_FAR, 0x10		/* LPT=378, COM1=3F8, COM2=2F8          */
	.byte PC97338_PTR, 0x00		/* Make sure COM1 test mode is cleared  */

serial_conf:
	.byte COM_MCR,     0x00		/* RTS/DTS off, disable loopback        */
	.byte COM_FCR,     0x07		/* Enable & reset FIFOs. DMA mode 0.    */
	.byte COM_LCR,     0x80		/* Set DLAB (access baudrate registers) */
	.byte COM_BRD_LO,  0x01		/* Baud Rate 115200 = 0x0001            */
	.byte COM_BRD_HI,  0x00
	.byte COM_LCR,     0x03		/* Unset DLAB. Set 8N1 mode             */

msg1:
	.string "\r\nThe VMware BIOS reached Diagnostic Code: 8"
     /* .string usefully adds a NUL terminator at the end of the text for us!   */
msg2:
	.string "h (in its BB.ROM file).\r\n"

/********************************************************************************/


/********************************************************************************/
/* reset: this section must reside at 0xfffffff0, and be exactly 16 bytes       */
/********************************************************************************/
.section reset, "ax"
	/* Issue a manual jmp to work around a binutils bug.                    */
	/* See coreboot's src/cpu/x86/16bit/reset16.inc                         */
	.byte  0xe9
	.int   init - ( . + 2 )
	.align 16, 0xff	/* fills section to end of ROM (with 0xFF)              */
/********************************************************************************/
